import 'package:analyzer/dart/element/element.dart';
import 'package:build/build.dart';
import 'package:fireproof/src/utils.dart';
import 'package:fireproof_annotation/fireproof_annotation.dart';
import 'package:source_gen/source_gen.dart';

class CollectionOutput {
  /// The string name of the collection the firestore database
  final String collectionName;

  final String collectionClassName;

  final String fieldName;

  final String output;

  final String docName;

  final String modelName;

  const CollectionOutput({
    required this.collectionName,
    required this.collectionClassName,
    required this.fieldName,
    required this.output,
    required this.docName,
    required this.modelName,
  });
}

Future<CollectionOutput> collectionGenerator(
  PropertyAccessorElement element,
  BuildStep buildStep,
) async {
  final collectionChecker = TypeChecker.fromRuntime(Collection);
  final annotation = collectionChecker.firstAnnotationOfExact(element)!;
  final stringBuffer = StringBuffer();
  final displayName = element.displayName;
  final collectionName =
      annotation.getField('name')?.toStringValue() ?? displayName;
  final node = await tryGetAstNodeForElement(element, buildStep);
  final astType = getGenericTypeTokens(node);
  final baseType = astType.single;
  final library = (await buildStep.inputLibrary)
      .importedLibraries
      .firstWhere((library) => library.getType(baseType.lexeme) != null);
  final originalType = library.getType(baseType.lexeme)!;
  final baseTypeConstructor = originalType.unnamedConstructor;

  if (baseTypeConstructor == null) {
    throw InvalidGenerationSourceError(
      'The Collection\'s generic type class must have a factory '
      'constructor element: ${element.name}',
      element: element,
    );
  }

  // TODO: FirstWhereOrNull
  final baseModelParameter = baseTypeConstructor.parameters
      .firstWhere((parameter) => parameter.displayName == 'data');
  final baseElement = baseModelParameter.type.element as ClassElement?;

  if (baseElement == null ||
      baseElement.getNamedConstructor('fromJson') == null) {
    throw InvalidGenerationSourceError(
      'The Collection\'s generic base type must have a fromJson factory '
      'constructor element: ${element.name}',
      element: element,
    );
  }

  final collectionToken = getTypeToken(node);

  if (collectionToken == null) {
    throw InvalidGenerationSourceError(
      'Could not find a valid type for element '
      'constructor element: ${element.name}',
      element: element,
    );
  }

  // RENAME ALL OF THIS
  final modelName = baseElement.displayName;
  final className = collectionToken.lexeme;
  final originalTypeName = originalType.displayName;
  final docName = originalTypeName + 'Doc';
  final lowerCaseModel = lowerTitleCase(originalTypeName);
  final querySnapshotName = originalTypeName + 'QuerySnapshot';
  final queryName = originalTypeName + 'Query';

  // TODO: How do I get the parent of this?

  stringBuffer.write('''
class $className<T extends $originalTypeName> extends $queryName with FireCollectionReference<$modelName, $docName> {
  final CollectionReference _reference;

  $className(this._reference) : super(_reference);

  @override
  FirebaseFirestore get firestore => _reference.firestore;

  @override
  String get id => _reference.id;

  @override
  FireCollectionReference? get parent => throw UnimplementedError();

  @override
  String get path => _reference.path;

  @override
  Future<$docName> add($modelName $lowerCaseModel) async {
    final docReference = await _reference.add($lowerCaseModel.toJson());

    return $docName(docReference);
  }

  @override
  $docName doc([String? path]) {
    final documentReference = _reference.doc(path);

    return $docName(documentReference);
  }

  @override
  Stream<$querySnapshotName> snapshots({bool includeMetadataChanges = false}) {
    return _reference
        .snapshots(includeMetadataChanges: includeMetadataChanges)
        .map((snapshot) => $querySnapshotName(snapshot));
  }

  @override
  Future<$querySnapshotName> get([GetOptions? getOptions]) async {
    final snapshot = await _reference.get(getOptions);

    return $querySnapshotName(snapshot);
  }
}

''');

  stringBuffer.write('''
class $queryName extends FireQuery<$originalTypeName, $docName> {
  final Query _query;

  $queryName(this._query);

  @override
  FirebaseFirestore get firestore => _query.firestore;

  @override
  Map<String, dynamic> get parameters {
    return _query.parameters;
  }

  @override
  $queryName endAtDocument(DocumentSnapshot documentSnapshot) {
    final query = _query.endAtDocument(documentSnapshot);

    return $queryName(query);
  }

  @override
  $queryName endAt(List values) {
    final query = _query.endAt(values);

    return $queryName(query);
  }

  @override
  $queryName endBeforeDocument(DocumentSnapshot documentSnapshot) {
    final query = _query.endBeforeDocument(documentSnapshot);

    return $queryName(query);
  }

  @override
  $queryName endBefore(List values) {
    final query = _query.endBefore(values);

    return $queryName(query);
  }

  @override
  Future<$querySnapshotName> get([GetOptions? options]) async {
    final querySnapshot = await _query.get(options);

    return $querySnapshotName(querySnapshot);
  }

  @override
  $queryName limit(int limit) {
    final query = _query.limit(limit);

    return $queryName(query);
  }

  @override
  $queryName limitToLast(int limit) {
    final query = _query.limit(limit);

    return $queryName(query);
  }

  @override
  Stream<$querySnapshotName> snapshots({
    bool includeMetadataChanges = false,
  }) async* {
    yield* _query.snapshots().map((snapshot) => $querySnapshotName(snapshot));
  }

  @override
  $queryName orderBy(dynamic field, {bool descending = false}) {
    final query = _query.orderBy(field, descending: descending);

    return $queryName(query);
  }

  @override
  $queryName startAfterDocument(DocumentSnapshot documentSnapshot) {
    // TODO: Typesafe FireDocumentSnapshot should be used
    final query = _query.startAfterDocument(documentSnapshot);

    return $queryName(query);
  }

  @override
  $queryName startAfter(List values) {
    final query = _query.startAfter(values);

    return $queryName(query);
  }

  @override
  $queryName startAtDocument(DocumentSnapshot documentSnapshot) {
    final query = _query.startAtDocument(documentSnapshot);

    return $queryName(query);
  }

  @override
  $queryName startAt(List values) {
    final query = _query.startAt(values);

    return $queryName(query);
  }

  @override
  $queryName where(dynamic field,
      {dynamic isEqualTo,
      dynamic isNotEqualTo,
      dynamic isLessThan,
      dynamic isLessThanOrEqualTo,
      dynamic isGreaterThan,
      dynamic isGreaterThanOrEqualTo,
      dynamic arrayContains,
      List? arrayContainsAny,
      List? whereIn,
      List? whereNotIn,
      bool? isNull}){
    return $queryName(_query.where(
      field,
      isEqualTo: isEqualTo,
      isNotEqualTo: isNotEqualTo,
      isLessThan: isLessThan,
      isLessThanOrEqualTo: isLessThanOrEqualTo,
      isGreaterThan: isGreaterThan,
      isGreaterThanOrEqualTo: isGreaterThanOrEqualTo,
      arrayContains: arrayContains,
      arrayContainsAny: arrayContainsAny,
      whereIn: whereIn,
      whereNotIn: whereNotIn,
      isNull: isNull,
    ));
  }
}
''');

  final queryDocumentSnapshotName = originalTypeName + 'QueryDocumentSnapshot';

  stringBuffer.write('''
class $querySnapshotName with FireQuerySnapshot<$originalTypeName, $docName> {
  final QuerySnapshot _snapshot;

  const $querySnapshotName(this._snapshot);

  @override
  List<$queryDocumentSnapshotName> get docs {
    return _snapshot.docs.map((document) {
      if (!document.exists) {
        throw StateError('Document \$document does not exist');
      }

      final data = $modelName.fromJson(document.data()!);

      return $queryDocumentSnapshotName(
        data: data,
        snapshot: document,
      );
    }).toList();
  }

  @override
  List<FireDocumentChange<$originalTypeName, $docName>> get docChanges {
    throw UnimplementedError();
  }

  @override
  SnapshotMetadata get metadata => _snapshot.metadata;
}

''');

  return CollectionOutput(
    fieldName: displayName,
    output: stringBuffer.toString(),
    collectionName: collectionName,
    collectionClassName: className,
    modelName: modelName,
    docName: docName,
  );
}
