import 'dart:async';

import 'package:analyzer/dart/element/element.dart';
import 'package:build/build.dart';
import 'package:fireproof/src/collection_generator.dart';
import 'package:fireproof_annotation/fireproof_annotation.dart'
    show Collection, Fireproof;
import 'package:source_gen/source_gen.dart';

class FireproofGenerator extends GeneratorForAnnotation<Fireproof> {
  @override
  FutureOr<String> generateForAnnotatedElement(
    Element element,
    ConstantReader annotation,
    BuildStep buildStep,
  ) async {
    if (element is! ClassElement) {
      throw InvalidGenerationSourceError(
        '@fireproof can only be applied on classes. Failing element: ${element.name}',
        element: element,
      );
    }

    final buffer = StringBuffer();
    final className = element.displayName;
    final prefix = '_';
    final mixinPrefix = prefix + '\$';
    final concreteClassName = prefix + className;
    final mixinName = mixinPrefix + className;
    final factoryFunctionName = mixinName + 'FromSnapshot';
    final constructor = element.unnamedConstructor;
    final privateConstructor = element.getNamedConstructor('_');
    final collectionChecker = TypeChecker.fromRuntime(Collection);
    final queryDocumentSnapshotName = className + 'QueryDocumentSnapshot';
    final collectionGetters = element.accessors
        .where((accessor) => accessor.isGetter)
        .where((element) => collectionChecker.hasAnnotationOf(element));

    if (!element.isAbstract && collectionGetters.isNotEmpty) {
      throw InvalidGenerationSourceError(
        '@fireproof classes that declare collections must be marked abstract. '
        'Failing element: ${element.name}',
        element: element,
      );
    }

    final collectionOutputFutures = collectionGetters
        .map<Future<CollectionOutput>>(
            (element) => collectionGenerator(element, buildStep))
        .toList();
    final collectionOutputs = await Future.wait(collectionOutputFutures);

    if (constructor == null || !constructor.isFactory) {
      throw InvalidGenerationSourceError(
        '@fireproof must have one unnamed factory constructor. Failing element:'
        ' ${element.name}',
        element: element,
      );
    }

    final dataParameter = constructor.parameters.singleWhere((element) {
      return element.displayName == 'data';
    }, orElse: () {
      throw InvalidGenerationSourceError(
        '@fireproof\'s factory must have a paramter named data. Failing'
        ' element: ${element.name}',
        element: element,
      );
    });

    final dataType =
        dataParameter.type.getDisplayString(withNullability: false);

    buffer.write('''
$className $factoryFunctionName(DocumentSnapshot snapshot) {
  final data = snapshot.data();

  if (!snapshot.exists) {
    throw StateError('Snapshot \$snapshot does not exist');
  }

  if (data == null) {
    throw StateError('No data for snapshot \$snapshot');
  }

  return $className(
    data: $dataType.fromJson(data),
    snapshot: snapshot,
  );
}

''');

    final docClassName = className + 'Doc';

    buffer.write('''
mixin $mixinName implements FireDocumentSnapshot<$dataType, $docClassName> {
  final $docClassName reference = throw StateError('This operation was not intended to be called on the object in its current state');

  @override
  final $dataType data = throw StateError('This operation was not intended to be called on the object in its current state');

  @override
  String get id => throw StateError('This operation was not intended to be called on the object in its current state');

  @override
  SnapshotMetadata get metadata => throw StateError('This operation was not intended to be called on the object in its current state');

  @override
  bool get exists => throw StateError('This operation was not intended to be called on the object in its current state');

  @override
  dynamic get(dynamic field) => throw StateError('This operation was not intended to be called on the object in its current state');
}

''');

    final namedParameters = constructor.parameters
        .where((element) => element.isNamed)
        .map((parameter) {
      var paramaterName = '';

      if (parameter.isNotOptional) {
        paramaterName = 'required ';
      }

      if (parameter.name == 'snapshot') {
        paramaterName += '${parameter.type} ${parameter.name}';
      } else {
        paramaterName += 'this.${parameter.name}';
      }

      return paramaterName;
    });
    final namedParametersString =
        namedParameters.isEmpty ? '' : '{${namedParameters.join(', ')}}';
    final optionalPositional = constructor.parameters
        .where((element) => element.isOptionalPositional)
        .map((parameter) {
      return 'this.${parameter.name}';
    });
    final optionalPositionalString =
        optionalPositional.isEmpty ? '' : '\[${optionalPositional.join(',')}\]';
    final requiredPositional = constructor.parameters
        .where((element) => element.isRequiredPositional)
        .map((parameter) {
      return 'this.${parameter.name}';
    });
    final requiredPositionalString =
        requiredPositional.isEmpty ? '' : requiredPositional.join(',');
    final constructorDeclaration = [
      requiredPositionalString,
      namedParametersString,
      optionalPositionalString,
    ].where((string) => string.isNotEmpty).join(',');
    final collectionFields = collectionOutputs.map((output) {
      return '''@override
final ${output.collectionClassName} ${output.fieldName};''';
    }).join('\n\n\t');
    final collectionInitializers = collectionOutputs.map((output) {
      return '${output.fieldName} = ${output.collectionClassName}(snapshot.reference.collection(\'${output.collectionName}\'))';
    }).join(',\n');
    final collectionInitializersString =
        collectionInitializers.isEmpty ? '' : ', $collectionInitializers';
    final inheritance = privateConstructor == null ? 'implements' : 'extends';
    final superString = privateConstructor == null ? '' : ', super._()';

    // Generates abstract FireDocumentSnapshot
    buffer.write('''
class $concreteClassName $inheritance $className {
  final DocumentSnapshot _snapshot;

  @override
  final $docClassName reference;

  $concreteClassName($constructorDeclaration)
    : _snapshot = snapshot,
      reference = $docClassName(snapshot.reference)
      $collectionInitializersString
      $superString;

  @override
  final $dataType data;

  @override
  String get id => reference.id;

  $collectionFields

  @override
  SnapshotMetadata get metadata => _snapshot.metadata;

  @override
  bool get exists => _snapshot.exists;

  @override
  dynamic get(dynamic field) {
    return _snapshot.get(field);
  }
}

''');

    // TODO: Calling collectionType should instead take a generated
    // enum of all currently defined sub collections types rather than a
    // string.

    // Generates the concrete document reference
    buffer.write('''
class $docClassName with FireDocumentReference<$dataType, $docClassName> {
  final DocumentReference _reference;

  $docClassName(this._reference);

  @override
  FirebaseFirestore get firestore => _reference.firestore;

  @override
  String get id => _reference.id;

  @override
  // FireCollectionReference get parent => _reference.parent;
  FireCollectionReference get parent => throw UnimplementedError();

  @override
  String get path => _reference.path;

  @override
  FireCollectionReference<$dataType, $docClassName> collection(String collectionPath) {
    throw UnimplementedError();
  }

  @override
  Future<void> delete() async {
    await _reference.delete();
  }

  @override
  Future<$queryDocumentSnapshotName> get([GetOptions? options]) async {
    final documentSnapshot = await _reference.get(options);

    if (!documentSnapshot.exists) {
      throw StateError('DocumentSnapshot \$documentSnapshot does not exist');
    }

    final data = $dataType.fromJson(documentSnapshot.data()!);

    return $queryDocumentSnapshotName(
      data: data,
      snapshot: documentSnapshot,
    );
  }

  @override
  Future<void> set($dataType data, [SetOptions? options]) async {
    await _reference.set(data.toJson(), options);
  }

  @override
  Stream<$queryDocumentSnapshotName> snapshots({bool includeMetadataChanges = false}) async* {
    yield* _reference
        .snapshots(includeMetadataChanges: includeMetadataChanges)
        .map(
      (documentSnapshot) {
        if (!documentSnapshot.exists) {
          throw StateError('DocumentSnapshot \$documentSnapshot does not exist');
        }

        final data = $dataType.fromJson(documentSnapshot.data()!);

        return $queryDocumentSnapshotName(
          data: data,
          snapshot: documentSnapshot,
        );
      },
    );
  }

  @override
  Future<void> update($dataType data) async {
    await _reference.update(data.toJson());
  }
}
''');

    buffer.write('''
class $queryDocumentSnapshotName extends $concreteClassName {
  $queryDocumentSnapshotName({
    required $dataType data,
    required DocumentSnapshot snapshot,
  }) : super(
    data: data,
    snapshot: snapshot,
  );

  @override
  bool get exists => true;
}
''');

    for (final output in collectionOutputs) {
      buffer.write(output.output);
    }

    return buffer.toString();
  }
}
